<?php

namespace App\domain\amadeus;

use App\domain\amadeus\traits\AmadeusDateConversionTrait;
use App\domain\amadeus\traits\AmadeusParamsTraits;
use App\Jobs\AmadeusCombinedOffersSavingJob;
use App\Traits\FileStorageTrait;
use Illuminate\Support\Collection;

class QueryCombinedOffersAction
{
    use FileStorageTrait;
    use AmadeusDateConversionTrait;
    use AmadeusParamsTraits;

    protected QueryFareMasterPricerTravelBoardSearchAction $queryFareMasterPricerAction;
    protected Collection $flightIndex;
    protected array $outboundGroupOfFlights;
    protected array $inboundGroupOfFlights;
    protected int $noOfLegs;
    protected string $requestId;

    public function __construct()
    {
        $this->queryFareMasterPricerAction = new QueryFareMasterPricerTravelBoardSearchAction();
    }

    public function execute(array $passengerCriteria, array $searchCriteriaFlight, ?array $searchModifiers): array
    {
        $this->requestId = uuid_create();

        $this->noOfLegs = count($searchCriteriaFlight);

        $flightOffers = $this->queryFareMasterPricerAction->execute($passengerCriteria, $searchCriteriaFlight, $searchModifiers);

        $combinedOffers = $this->getCombinedOffers($flightOffers[0]['Fare_MasterPricerTravelBoardSearchReply']);

        self::storeInFile("json", "0-flight-search-combined-offers", $combinedOffers->toJson());

        $combinedOffers = $combinedOffers
            ->sortBy(fn($combination) => $combination[0]['bestCombinablePrice']['TotalPrice']);

        dispatch(new AmadeusCombinedOffersSavingJob($this->requestId, $combinedOffers));

        return [
            'amaTotalCombinedOffers' => $combinedOffers->count(),
            'amaOffersRequestId' => $this->requestId,
            'productOffers' => $combinedOffers->take($this->getOffersPerPage())->toArray()
        ];
    }

    protected function getCombinedOffers($fare_MasterPricerTravelBoardSearchReply): Collection
    {

        $flightIndex = Collection::wrap($fare_MasterPricerTravelBoardSearchReply['flightIndex']);

        $this->separateOutboundAndInboundOffers($flightIndex);

        $combinedOffers = $this->combinedAllLegsInSingleJourney($fare_MasterPricerTravelBoardSearchReply['recommendation']);

        return $combinedOffers->flatten(1)->groupBy("combinabilityCode");
    }

    protected function combinedAllLegsInSingleJourney($recommendations): Collection
    {
        $combinedOffers = new Collection;

        foreach ($recommendations as $recommendation) {

            for ($leg = 0; $leg < $this->noOfLegs; $leg++) {
                $outboundOffer = $this->extractAndCombineSingleDirectionOffersSet(
                    $recommendation,
                    $leg
                );

                $combinedOffers->add([
                    ...$outboundOffer,
                ]);
            }
        }
        return $combinedOffers;
    }

    protected function separateOutboundAndInboundOffers(Collection $flightIndex): void
    {
        if (isset($flightIndex[0])) {
            $this->outboundGroupOfFlights = $flightIndex->firstWhere('requestedSegmentRef.segRef', 1)['groupOfFlights'];
            $this->inboundGroupOfFlights = $flightIndex->firstWhere('requestedSegmentRef.segRef', 2)['groupOfFlights'];

        } else {
            $this->outboundGroupOfFlights = $flightIndex->toArray()['groupOfFlights'];
            $this->inboundGroupOfFlights = [];
        }
    }

    protected function extractAndCombineSingleDirectionOffersSet(
        $recommendation,
        $legIndex
    ): array
    {
        $combinableOffers = [];

        if (isset($recommendation["segmentFlightRef"]["referencingDetail"])) {
            $combinableOffers[] = $this->extractSingleLegOffer($recommendation, $recommendation["segmentFlightRef"]["referencingDetail"], $legIndex);
        } else {
            foreach ($recommendation['segmentFlightRef'] as $segmentRef) {
                $combinableOffers[] = $this->extractSingleLegOffer($recommendation, $segmentRef["referencingDetail"], $legIndex);
            }
        }

//        self::storeInFile("json", "0-duplicate-offers", json_encode($combinableOffers));

        $uniqueOffers = $this->removedDuplicateOffers($combinableOffers);

//        self::storeInFile("json", "0-unique-offers", json_encode($uniqueOffers));
//
//
//        dd($uniqueOffers);

        return $uniqueOffers;
    }

    protected function extractSingleLegOffer($recommendation, $referencingDetail, $legIndex): array
    {
        $combinabilityCode = $recommendation['itemNumber']['itemNumberId']['number'];

        $productRefNo = $this->getProductRefNo($referencingDetail, $legIndex);

        $groupOfFlight = $this->findGroupOfFlight($productRefNo, $legIndex);
        $oneWaySegments = $groupOfFlight['flightDetails'];


        if (isset($oneWaySegments[0])) {
            $firstSegment = $oneWaySegments[0];
            $lastSegment = $oneWaySegments[count($oneWaySegments) - 1];
        } else {
            $firstSegment = $lastSegment = $oneWaySegments;
        }

        $oneWayPrice = $recommendation['recPriceInfo']['monetaryDetail'][0]['amount'];

//        $totalPrice = 0;
//        foreach ($recommendation['recPriceInfo']['monetaryDetail'] as $monetary) {
//            $totalPrice += floatval($monetary['amount']);
//        }

        $totalPrice = $recommendation['recPriceInfo']['monetaryDetail'][0]['amount'];

        $originDepartureTime = $this->getCarbonDateTime(
            $firstSegment['flightInformation']['productDateTime']['dateOfDeparture'],
            $firstSegment['flightInformation']['productDateTime']['timeOfDeparture'],
        );

        $destinationArrivalTime = $this->getCarbonDateTime(
            $lastSegment['flightInformation']['productDateTime']['dateOfArrival'],
            $lastSegment['flightInformation']['productDateTime']['timeOfArrival'],
        );

        $durationInHours = $this->getDurationInHours($originDepartureTime, $destinationArrivalTime);

        return [
            '@type' => $legIndex == 0 ? 'Outbound' : 'Inbound',
            'combinabilityCode' => $combinabilityCode,
            'catalogProductOfferingsIdentifier' => "",
            'authority' => "AMA-GDS",
            'offerId' => $productRefNo,
            'id' => $productRefNo,
            'fromAirport' => $firstSegment['flightInformation']['location'][0]['locationId'],
            'toAirport' => $lastSegment['flightInformation']['location'][1]['locationId'],
            'price' => [
                'TotalPrice' => $oneWayPrice,
                'CurrencyCode' => [
                    'value' => "GBP"
                ]
            ],
            'bestCombinablePrice' => [
                'TotalPrice' => $totalPrice,
                'CurrencyCode' => [
                    'value' => "GBP"
                ]
            ],
            'productAir' => $recommendation,
            'fromAirportCarrierCode' => $firstSegment['flightInformation']['companyId']['marketingCarrier'],
            'fromAirportAircraftCode' => $firstSegment['flightInformation']['flightOrtrainNumber'],
            'toAirportCarrierCode' => $lastSegment['flightInformation']['companyId']['marketingCarrier'],
            'toAirportAircraftCode' => $lastSegment['flightInformation']['flightOrtrainNumber'],
            'originDepartureTime' => $originDepartureTime,
            'destinationArrivalTime' => $destinationArrivalTime,
            'duration' => $this->getDurationString($durationInHours),
            'numberOfStops' => count($oneWaySegments) - 1,
            'segments' => $groupOfFlight,
            'durationInHours' => $durationInHours,
            'numberOfBookableSeats' => $productIAir['Quantity'] ?? 0,
        ];
    }

    protected function getProductRefNo(array $referencingDetail, int $legIndex)
    {
        $refKeys = Collection::wrap($referencingDetail)
            ->where('refQualifier', 'S');

        return $refKeys[$legIndex]['refNumber'];
    }

    protected function findGroupOfFlight(string $refNo, int $legIndex)
    {
        return ($legIndex === 0)
            ? Collection::wrap($this->outboundGroupOfFlights)
                ->firstWhere('propFlightGrDetail.flightProposal.0.ref', $refNo)
            : Collection::wrap($this->inboundGroupOfFlights)
                ->firstWhere('propFlightGrDetail.flightProposal.0.ref', $refNo);
    }

    protected function removedDuplicateOffers($flights): array
    {
        $uniqueFlights = [];

        foreach ($flights as $flight) {
            $key = $flight['@type'] . '-' . $flight['id'];

            if (!isset($uniqueFlights[$key])) {
                // If not, add it
                $uniqueFlights[$key] = $flight;
            }
        }

        return array_values($uniqueFlights);


    }
}
